from typing import List, Union
from abstra_internals.widgets.widget_base import AbstraOption, Input, MultipleHandler, OptionsHandler


class MultipleChoiceInput(Input):
    type = 'multiple-choice-input'

    def __init__(self, key: str, label: str, options: List[AbstraOption],
        **kwargs):
        super().__init__(key)
        self.set_props(dict(label=label, options=options, **kwargs))

    def set_props(self, props):
        self.label = props.get('label', 'Label')
        self.options = props.get('options', [])
        self.required = props.get('required', True)
        self.hint = props.get('hint', None)
        self.full_width = props.get('full_width', False)
        self.min = props.get('min', None)
        self.max = props.get('max', None)
        self.disabled = props.get('disabled', False)
        self.multiple = props.get('multiple', False)
        self.empty_value = [] if self.multiple else None
        self.value = props.get('initial_value', self.empty_value)
        self.multiple_handler = MultipleHandler(self.multiple, self.min,
            self.max, self.required)
        self.options_handler = OptionsHandler(self.options)

    def render(self, ctx: dict):
        return {'type': self.type, 'key': self.key, 'label': self.label,
            'options': self.options_handler.serialized_options(), 'hint':
            self.hint, 'multiple': self.multiple, 'value': self.
            serialize_value(), 'required': self.required and not self.
            multiple, 'fullWidth': self.full_width, 'min': self.min, 'max':
            self.max, 'disabled': self.disabled, 'errors': self.errors}

    def validate(self) ->List[str]:
        return self.multiple_handler.validate(self.value)

    def serialize_value(self) ->List:
        list_values = self.multiple_handler.value_to_list(self.value)
        return self.options_handler.ids_from_values(list_values)

    def parse_value(self, value: Union[List[str], None]):
        list_values = self.options_handler.values_from_ids(value)
        return self.multiple_handler.value_to_list_or_value(list_values)



import json
from dataclasses import dataclass
from typing import Any, Dict, List

import flask

from abstra_internals.cloud_api import (
    cancel_all,
    create_thread,
    generate_project,
    get_ai_messages,
)
from abstra_internals.credentials import resolve_headers
from abstra_internals.repositories.project.json_migrations import get_latest_version
from abstra_internals.repositories.project.project import Project, ProjectRepository
from abstra_internals.server.controller.linters import fix_all_linters
from abstra_internals.server.controller.main import MainController
from abstra_internals.settings import Settings
from abstra_internals.usage import usage


@dataclass
class PythonFile:
    filename: str
    content: str
    stage: str

    @staticmethod
    def from_dict(data: Dict[str, Any]) -> "PythonFile":
        return PythonFile(data["filename"], data["content"], data["stage"])


class AiController:
    def __init__(self, controller: MainController):
        self.controller = controller

    def send_ai_message(self, messages, stage, thread_id):
        headers = resolve_headers() or {}
        yield from get_ai_messages(messages, stage, thread_id, headers)

    def create_thread(self):
        headers = resolve_headers()
        if headers is None:
            return None
        return create_thread(headers)

    def cancel_all(self, thread_id: str):
        headers = resolve_headers()
        if headers is None:
            return None
        return cancel_all(headers, thread_id)

    def generate_project(self, prompt: str, tries: int = 0):
        headers = resolve_headers() or {}
        abstra_json_version = get_latest_version()
        try:
            res = generate_project(prompt, abstra_json_version, headers)
            is_abstra_json_valid = Project.validate(res["abstra_json"])
            if not is_abstra_json_valid:
                raise Exception("Generated abstra.json is not valid")

            generated_abstra_json = json.dumps(res["abstra_json"], indent=4)
            python_files = [PythonFile.from_dict(file) for file in res["python_files"]]

            self.init_stages(python_files)
            Settings.root_path.joinpath("abstra.json").write_text(
                generated_abstra_json, encoding="utf-8"
            )

            ProjectRepository.initialize_or_migrate(verbose=False)
            fix_all_linters()
        except Exception as e:
            if tries >= 3:
                raise e
            return self.generate_project(prompt, tries + 1)

    def init_stages(self, python_files: List[PythonFile]):
        for file in python_files:
            if file.stage == "form":
                script = self.controller.create_form(file.filename[:-3], file.filename)
            elif file.stage == "hook":
                script = self.controller.create_hook(file.filename[:-3], file.filename)
            elif file.stage == "script":
                script = self.controller.create_script(
                    file.filename[:-3], file.filename
                )
            elif file.stage == "job":
                script = self.controller.create_job(file.filename[:-3], file.filename)
            else:
                raise Exception(f"Invalid stage {file.stage}")
            script.file_path.write_text(file.content, encoding="utf-8")


def get_editor_bp(main_controller: MainController):
    bp = flask.Blueprint("editor_ai", __name__)
    controller = AiController(main_controller)

    @bp.post("/message")
    @usage
    def _get_next_message():
        body = flask.request.json
        if not body:
            flask.abort(400)

        streamer = controller.send_ai_message(
            body["messages"], body["runtime"], body["threadId"]
        )

        if streamer is None:
            flask.abort(403)

        return flask.Response(streamer, mimetype="text/event-stream")

    @bp.post("/thread")
    @usage
    def _create_thread():
        thread = controller.create_thread()
        if not thread:
            flask.abort(403)
        return thread

    @bp.post("/cancel-all")
    @usage
    def _cancel_all():
        body = flask.request.json
        if not body:
            flask.abort(400)
        thread_id = body.get("threadId")
        if not thread_id:
            flask.abort(400)
        thread = controller.cancel_all(thread_id)
        if not thread:
            flask.abort(403)
        return thread

    @bp.post("/generate")
    @usage
    def _generate():
        body = flask.request.json
        if not body:
            flask.abort(400)

        prompt = body.get("prompt")
        if not prompt:
            flask.abort(400)

        controller.generate_project(prompt)
        return {"success": True}

    @bp.post("/vote")
    def _vote():
        body = flask.request.json
        if not body:
            flask.abort(400)
        vote = body.get("vote")
        question = body.get("question")
        answer = body.get("answer")
        context = body.get("context")

        @usage
        def ai_vote(vote, question, answer, context):
            return

        ai_vote(vote, question, answer, context)
        return {"success": True}

    return bp



from abstra_internals.widgets.widget_base import Input


class ToggleInput(Input):
    type = 'toggle-input'
    empty_value: bool = False

    def __init__(self, key: str, label: str, **kwargs):
        super().__init__(key)
        self.set_props(dict(label=label, **kwargs))

    def set_props(self, props):
        self.label = props.get('label', 'Label')
        self.on_text = props.get('on_text', 'Yes')
        self.off_text = props.get('off_text', 'No')
        self.value = props.get('initial_value', False)
        self.required = props.get('required', True)
        self.hint = props.get('hint', None)
        self.full_width = props.get('full_width', False)
        self.disabled = props.get('disabled', False)

    def is_value_unset(self):
        return False

    def render(self, ctx: dict):
        return {'type': self.type, 'key': self.key, 'label': self.label,
            'onText': self.on_text, 'offText': self.off_text, 'value': self
            .serialize_value(), 'required': self.required, 'hint': self.
            hint, 'fullWidth': self.full_width, 'disabled': self.disabled,
            'errors': self.errors}

    def serialize_value(self) ->bool:
        return self.value or False

    def parse_value(self, value: bool) ->bool:
        return bool(value)



import unittest

from abstra_internals.repositories.stage_run import GetStageRunByQueryFilter
from abstra_internals.utils.filter import FilterCondition


class TestGetStageRunByQueryFilter(unittest.TestCase):
    def test_to_dict(self):
        filter = GetStageRunByQueryFilter(
            stage="stage_id",
            assignee=None,
            parent_id=None,
            status=["waiting"],
            data=None,
            search=None,
            advanced_data_filter=FilterCondition.from_dict(
                {
                    "operator": "AND",
                    "conditions": [
                        {"key": "key1", "comparator": "is", "value": "value1"},
                        {"key": "key2", "comparator": "is", "value": "value2"},
                    ],
                }
            ),
        )
        self.assertEqual(
            filter.to_dict(),
            {
                "stage": "stage_id",
                "status": ["waiting"],
                "advanced_data_filter": '{"operator": "AND", "conditions": [{"key": "key1", "comparator": "is", "value": "value1"}, {"key": "key2", "comparator": "is", "value": "value2"}]}',
            },
        )



import json
from io import BytesIO
from typing import Dict, List, Optional, Tuple, Union

from abstra_internals.contract.forms import ValidationResult
from abstra_internals.controllers.execution_client_hook import HookClient
from abstra_internals.controllers.execution_store import (
    ExecutionNotFound,
    ExecutionStore,
)
from abstra_internals.interface.sdk import user_exceptions
from abstra_internals.jwt_auth import UserClaims
from abstra_internals.repositories.execution import ExecutionRepository
from abstra_internals.repositories.project.project import ProjectRepository
from abstra_internals.repositories.stage_run import StageRun, StageRunRepository
from abstra_internals.utils import is_json_serializable
from abstra_internals.utils.insensitive_dict import CaseInsensitiveDict
from abstra_internals.utils.json import to_json_serializable


class HookSDKController:
    def get_current_client(self) -> HookClient:
        try:
            return ExecutionStore.get_hook_client()
        except ExecutionNotFound:
            raise user_exceptions.UnboundSDK()

    def set_response(self, status: int, body: str, headers: Dict[str, str]):
        self.get_current_client().set_response(status, body, headers)

    def get_raw_request(self) -> Tuple[str, Dict[str, str], CaseInsensitiveDict]:
        request = self.get_current_client().get_request()
        return (
            request["body"],
            request["query_params"],
            CaseInsensitiveDict(request["headers"]),
        )

    def get_request(
        self,
    ) -> Tuple[Union[str, List, Dict], Dict[str, str], CaseInsensitiveDict]:
        body, query, headers = self.get_raw_request()
        try:
            if "application/json" in headers["Content-Type"]:
                return json.loads(body), query, headers
            elif "multipart/form-data" in headers["Content-Type"]:
                return self._multipart_form_parse(body, headers), query, headers
            else:
                return body, query, headers
        except Exception:
            return body, query, headers

    def _multipart_form_parse(self, body: str, headers: dict):
        from multipart import MultipartParser, parse_options_header, to_bytes

        _, options = parse_options_header(headers["Content-Type"])
        boundary = options["boundary"].encode("utf-8")
        p = MultipartParser(BytesIO(to_bytes(body)), boundary)  # TODO: direct from fs
        return [{"name": i.name, "value": i.value} for i in p]


class FormSDKController:
    def __init__(self) -> None:
        try:
            self.client = ExecutionStore.get_form_client()
        except ExecutionNotFound:
            raise user_exceptions.UnboundSDK()

    def get_user(self, force_refresh: bool) -> Optional[UserClaims]:
        data = self.client.request_auth(force_refresh)
        claims = UserClaims.from_jwt(data["jwt"])

        if not claims:
            self.client.handle_invalid_jwt()
            return None

        self.client.handle_valid_jwt()

        return claims

    def execute_js(self, code: str, context: dict):
        return self.client.request_execute_js(code, context)

    def redirect(self, url: str, query_params: Optional[Dict[str, str]] = None) -> None:
        self.client.request_redirect(url, query_params)

    def get_query_params(self) -> Dict[str, str]:
        return self.client.get_query_params()

    def request_mount_page(
        self,
        widgets: list,
        actions: list,
        end_program: bool,
        reactive_polling_interval: int,
        steps_info: Optional[List],
    ):
        self.client.request_mount_page(
            widgets,
            actions,
            end_program,
            reactive_polling_interval,
            steps_info,
        )

    def request_page_update(
        self, widgets: list, validation: ValidationResult, event_seq: int
    ):
        self.client.request_page_update(widgets, validation, event_seq)

    def next_message(self) -> Dict:
        return self.client.next_message()


class WorkflowSDKController:
    def __init__(
        self,
        stage_run_repository: StageRunRepository,
        execution_repository: ExecutionRepository,
        project_repository: ProjectRepository,
    ) -> None:
        self.stage_run_repo = stage_run_repository
        self.execution_repo = execution_repository
        self.project_repo = project_repository
        self.project = project_repository.load()

    def _get_current_stage_run(self) -> StageRun:
        try:
            execution = ExecutionStore.get_by_thread().execution
            return self.stage_run_repo.get(execution.stage_run_id)
        except ExecutionNotFound:
            raise user_exceptions.UnboundSDK()

    def get_data(self, key: Optional[str]) -> Union[Dict[str, object], object]:
        stage_run = self._get_current_stage_run()
        return stage_run.data if not key else stage_run.data.get(key)

    def set_data(self, key: str, value: object) -> None:
        value = to_json_serializable(value)
        if not is_json_serializable(value):
            raise TypeError()

        stage_run = self._get_current_stage_run()

        stage_run.data[key] = value
        self.stage_run_repo.update_data(stage_run.id, stage_run.data)



from dataclasses import dataclass
from functools import wraps
from typing import Callable, Optional, Protocol

import flask


class CacheControl(Protocol):
    def __str__(self) -> str: ...


@dataclass
class NoCache:
    def __str__(self):
        return ""


@dataclass
class PublicCache:
    max_age: Optional[int] = None

    def __str__(self):
        if self.max_age is None:
            return "public"
        return f"public, max-age={self.max_age}"


@dataclass
class CacheCommand:
    cache_control: CacheControl
    proxy_cache_control: CacheControl


CachePolicy = Callable[[flask.Request, flask.Response], CacheCommand]


def cache_control(policy: CachePolicy):
    def decorator(f):
        @wraps(f)
        def decorated_function(*args, **kwargs):
            @flask.after_this_request
            def set_cache_control_header(response: flask.Response):
                cmd = policy(flask.request, response)
                if cmd.cache_control:
                    response.headers["Cache-Control"] = str(cmd.cache_control)
                if cmd.proxy_cache_control:
                    response.headers["X-Proxy-Cache-Control"] = str(
                        cmd.proxy_cache_control
                    )
                return response

            return f(*args, **kwargs)

        return decorated_function

    return decorator


class Cache:
    def __init__(self, enabled: bool):
        self.enabled = enabled

    def __assets_policy(self, _: flask.Request, __: flask.Response) -> CacheCommand:
        if not self.enabled:
            return CacheCommand(NoCache(), NoCache())

        return CacheCommand(PublicCache(1800), PublicCache())

    def __statics_policy(
        self, request: flask.Request, __: flask.Response
    ) -> CacheCommand:
        if not self.enabled:
            return CacheCommand(NoCache(), NoCache())

        if request.content_type == "text/html":
            return CacheCommand(NoCache(), PublicCache())

        return CacheCommand(PublicCache(86400), PublicCache())

    def assets(self):
        return cache_control(self.__assets_policy)

    def statics(self):
        return cache_control(self.__statics_policy)



from typing import List

from abstra_internals.email_templates.i18n import get_translation
from abstra_internals.repositories.email import EmailParams
from abstra_internals.repositories.project.project import ProjectRepository


def generate_email(to: List[str], message: str, title: str) -> EmailParams:
    project = ProjectRepository.load()
    translations = get_translation(project.workspace.language or "en")

    subject = translations.message_from(project.workspace.email_name)
    if title:
        subject += f": {title}"

    return EmailParams(
        kind="message",
        to=to,
        subject=subject,
        body=message,
        is_html=False,
    )



from unittest import TestCase

from abstra_internals.repositories.execution_logs import LocalExecutionLogsRepository
from abstra_internals.repositories.project.project import (
    FormStage,
    JobStage,
    NotificationTrigger,
    ProjectRepository,
)
from abstra_internals.repositories.stage_run import (
    CountStageRunsByStatus,
    LocalStageRunRepository,
)
from abstra_internals.server.controller.kanban import (
    DataRequest,
    KanbanController,
    KanbanData,
    StageCardContentItem,
    StageRunCard,
)
from tests.fixtures import clear_dir, init_dir


class KanbanTests(TestCase):
    def setUp(self):
        self.root = init_dir()

        self.stage_run_repository = LocalStageRunRepository()
        self.controller = KanbanController(
            execution_logs_repository=LocalExecutionLogsRepository(),
            stage_run_repository=self.stage_run_repository,
        )

    def tearDown(self) -> None:
        clear_dir(self.root)

    def test_get_data_empty(self):
        data = self.controller.get_data(
            DataRequest.from_dict(
                {
                    "selection": [],
                    "filter": {},
                }
            )
        )

        self.assertEqual(
            data, KanbanData(stage_run_cards=[], not_found_stages=[], total_count=0)
        )

    def test_get_data_stage_no_filters(self):
        project = ProjectRepository.load()

        job = JobStage(
            id="job",
            title="Job",
            file="job.py",
            schedule="* * * * *",
            workflow_position=(0, 0),
            workflow_transitions=[],
        )
        form = FormStage(
            id="form",
            path="form",
            title="Form",
            file="form.py",
            workflow_position=(0, 0),
            workflow_transitions=[],
            notification_trigger=NotificationTrigger(
                variable_name="val", enabled=False
            ),
        )
        project.add_stage(job)
        project.add_stage(form)
        ProjectRepository.save(project)
        job_stage_run = self.stage_run_repository.create_initial("job")
        form_stage_run = self.stage_run_repository.create_initial("form")

        data = self.controller.get_data(
            DataRequest.from_dict(
                {
                    "selection": [],
                    "filter": {},
                }
            )
        )

        self.assertEqual(
            data.to_dict(),
            KanbanData(
                stage_run_cards=[
                    StageRunCard(
                        form_stage_run.id,
                        created_at=form_stage_run.created_at,
                        updated_at=form_stage_run.updated_at,
                        status=form_stage_run.status,
                        content=[],
                        stage=form_stage_run.stage,
                    ),
                    StageRunCard(
                        job_stage_run.id,
                        created_at=job_stage_run.created_at,
                        updated_at=job_stage_run.updated_at,
                        status=job_stage_run.status,
                        content=[],
                        stage=job_stage_run.stage,
                    ),
                ],
                not_found_stages=[],
                total_count=2,
            ).to_dict(),
        )

    def test_get_data_with_stage_filter(self):
        self.maxDiff = None
        project = ProjectRepository.load()

        job = JobStage(
            id="job",
            title="Job",
            file="job.py",
            schedule="* * * * *",
            workflow_position=(0, 0),
            workflow_transitions=[],
        )
        form = FormStage(
            id="form",
            path="form",
            title="Form",
            file="form.py",
            workflow_position=(0, 0),
            workflow_transitions=[],
            notification_trigger=NotificationTrigger(
                variable_name="val", enabled=False
            ),
        )
        project.add_stage(job)
        project.add_stage(form)
        ProjectRepository.save(project)
        job_stage_run = self.stage_run_repository.create_initial("job")
        self.stage_run_repository.create_initial("form")

        data = self.controller.get_data(
            DataRequest.from_dict(
                {
                    "limit": 10,
                    "offset": 0,
                    "filter": {
                        "stage": ["job"],
                    },
                }
            )
        )

        self.assertEqual(
            data.to_dict(),
            KanbanData(
                stage_run_cards=[
                    StageRunCard(
                        job_stage_run.id,
                        created_at=job_stage_run.created_at,
                        updated_at=job_stage_run.updated_at,
                        status=job_stage_run.status,
                        content=[],
                        stage=job_stage_run.stage,
                    ),
                ],
                not_found_stages=[],
                total_count=1,
            ).to_dict(),
        )

    def test_get_data_with_advanced_data_filter(self):
        project = ProjectRepository.load()

        job = JobStage(
            id="job",
            title="Job",
            file="job.py",
            schedule="* * * * *",
            workflow_position=(0, 0),
            workflow_transitions=[],
        )
        project.add_stage(job)
        ProjectRepository.save(project)
        job_stage_run_1 = self.stage_run_repository.create_initial(
            "job", {"foo": "bar"}
        )
        job_stage_run_2 = self.stage_run_repository.create_initial(
            "job", {"foo": "contains bar"}
        )
        self.stage_run_repository.create_initial("job", {"foo": "baz"})

        data = self.controller.get_data(
            DataRequest.from_dict(
                {
                    "limit": 10,
                    "offset": 0,
                    "filter": {
                        "advanced_data_filter": {
                            "key": "foo",
                            "comparator": "contains",
                            "value": "bar",
                        },
                    },
                }
            )
        )

        self.assertEqual(
            data.to_dict(),
            KanbanData(
                stage_run_cards=[
                    StageRunCard(
                        job_stage_run_2.id,
                        created_at=job_stage_run_2.created_at,
                        updated_at=job_stage_run_2.updated_at,
                        status=job_stage_run_2.status,
                        content=[StageCardContentItem(key="foo", value="contains bar")],
                        stage=job_stage_run_2.stage,
                    ),
                    StageRunCard(
                        job_stage_run_1.id,
                        created_at=job_stage_run_1.created_at,
                        updated_at=job_stage_run_1.updated_at,
                        status=job_stage_run_1.status,
                        content=[StageCardContentItem(key="foo", value="bar")],
                        stage=job_stage_run_1.stage,
                    ),
                ],
                not_found_stages=[],
                total_count=2,
            ).to_dict(),
        )

    def test_get_job(self):
        project = ProjectRepository.load()

        job = JobStage(
            id="job",
            title="Job",
            file="job.py",
            schedule="* * * * *",
            workflow_position=(0, 0),
            workflow_transitions=[],
        )
        project.add_stage(job)
        ProjectRepository.save(project)

        rtvd = self.controller.get_job("job")

        self.assertEqual(rtvd, job)

    def test_count_by_status_one(self):
        project = ProjectRepository.load()

        form = FormStage(
            id="form",
            path="form",
            title="Form",
            file="form.py",
            workflow_position=(0, 0),
            workflow_transitions=[],
            notification_trigger=NotificationTrigger(
                variable_name="val", enabled=False
            ),
        )
        project.add_stage(form)
        ProjectRepository.save(project)
        self.stage_run_repository.create_initial("form")

        count = self.controller.count_by_status()

        self.assertEqual(count, [CountStageRunsByStatus("form", 1, "waiting")])

        count = self.controller.count_by_status()

        self.assertEqual(count, [CountStageRunsByStatus("form", 1, "waiting")])

    def test_count_by_status_multiple(self):
        project = ProjectRepository.load()

        form = FormStage(
            id="form",
            path="form",
            title="Form",
            file="form.py",
            workflow_position=(0, 0),
            workflow_transitions=[],
            notification_trigger=NotificationTrigger(
                variable_name="val", enabled=False
            ),
        )
        job = JobStage(
            id="job",
            title="Job",
            file="job.py",
            schedule="* * * * *",
            workflow_position=(0, 0),
            workflow_transitions=[],
        )
        project.add_stage(form)
        project.add_stage(job)
        ProjectRepository.save(project)
        form_sr1 = self.stage_run_repository.create_initial("form")
        form_sr2 = self.stage_run_repository.create_initial("form")
        form_sr3 = self.stage_run_repository.create_initial("form")
        form_sr4 = self.stage_run_repository.create_initial("form")
        job_sr1 = self.stage_run_repository.create_initial("job")
        _job_sr2 = self.stage_run_repository.create_initial("job")

        self.stage_run_repository.change_status(form_sr1.id, "running")
        self.stage_run_repository.change_status(form_sr2.id, "running")
        self.stage_run_repository.change_status(form_sr3.id, "running")
        self.stage_run_repository.change_status(form_sr4.id, "running")
        self.stage_run_repository.change_status(job_sr1.id, "running")

        self.stage_run_repository.change_status(form_sr3.id, "finished")
        self.stage_run_repository.change_status(form_sr4.id, "failed")

        count = self.controller.count_by_status()

        expected = [
            CountStageRunsByStatus("form", 2, "running"),
            CountStageRunsByStatus("form", 1, "finished"),
            CountStageRunsByStatus("form", 1, "failed"),
            CountStageRunsByStatus("job", 1, "waiting"),
            CountStageRunsByStatus("job", 1, "running"),
        ]

        self.assertCountEqual(count, expected)



from enum import Enum

import pkg_resources
import requests

from abstra_internals.environment import IS_PRODUCTION
from abstra_internals.settings import Settings
from abstra_internals.utils import packages as pkg_utils

VersionStatus = Enum(
    "VersionStatus", ["UP_TO_DATE", "OUT_OF_DATE", "LATEST_IS_OUTDATED", "UNKNOWN"]
)


class PackageVersionManager:
    def __init__(self, package_name: str) -> None:
        self.package_name = package_name
        self.package_display_name = (
            package_name if package_name != "abstra" else "Abstra Editor"
        )
        self.current_local_version = pkg_utils.get_local_package_version(package_name)
        self.cached_latest_version = (
            pkg_utils.get_cached_latest_version(Settings.root_path, package_name)
            or self.current_local_version
        )

    def get_version_status(self) -> VersionStatus:
        if IS_PRODUCTION:
            return VersionStatus.UNKNOWN

        try:
            if self.current_local_version < self.cached_latest_version:
                return VersionStatus.OUT_OF_DATE

            if self.current_local_version > self.cached_latest_version:
                pkg_utils.update_cached_latest_version(
                    Settings.root_path, self.current_local_version, self.package_name
                )
                return VersionStatus.LATEST_IS_OUTDATED

            return VersionStatus.UP_TO_DATE

        except pkg_resources.DistributionNotFound:
            print(f"{self.package_name} is not installed.\n")
            return VersionStatus.UNKNOWN

        except requests.exceptions.RequestException:
            print(
                f"Unable to fetch version information for {self.package_name} from PyPI.\n"
            )
            return VersionStatus.UNKNOWN

    def get_status_message(self) -> str:
        status = self.get_version_status()

        if (
            status == VersionStatus.UP_TO_DATE
            or status == VersionStatus.LATEST_IS_OUTDATED
        ):
            return f"{self.package_display_name} is up to date (version {self.current_local_version}).\n"
        elif status == VersionStatus.OUT_OF_DATE:
            return f"A new version of {self.package_display_name} is available. Latest version is {self.cached_latest_version}, but you have {self.current_local_version}.\nPlease run 'pip install {self.package_name} --upgrade' to update.\n"
        else:
            return ""

    def print_status_message(self):
        print(self.get_status_message())


def check_latest_version(package_name="abstra"):
    package_version = PackageVersionManager(package_name)
    package_version.print_status_message()



import sys
from typing import Literal

import pkg_resources
import sentry_sdk

from abstra_internals.compatibility.compat_traceback import print_exception
from abstra_internals.utils import is_dev_env, is_test_env


class DevSDK:
    @classmethod
    def init(cls, *args, **kwargs):
        pass

    @classmethod
    def capture_exception(cls, exception: Exception):
        print("[ABSTRA_LOGGER] Exception captured:", file=sys.stderr)
        print_exception(exception)

    @classmethod
    def capture_message(cls, message):
        print(f"[ABSTRA_LOGGER] Message captured: {message}")

    @classmethod
    def flush(cls):
        pass


class AbstraLogger:
    @classmethod
    def init(cls, environment: Literal["cloud", "local"]):
        try:
            cls.get_sdk().init(
                dsn="https://9bbccd1a46ddb8a563483c6afc61ca35@o1317386.ingest.us.sentry.io/4507024713383936",
                traces_sample_rate=1.0,
                profiles_sample_rate=1.0,
                environment=environment,
                enable_tracing=True,
                release=pkg_resources.get_distribution("abstra").version,
                shutdown_timeout=0,
            )
        except Exception:
            print("[ABSTRA_LOGGER]Error reporting has been turned off.")

    @classmethod
    def capture_exception(cls, exception: Exception):
        cls.get_sdk().capture_exception(exception)
        cls.get_sdk().flush()

    @classmethod
    def capture_message(cls, message: str):
        cls.get_sdk().capture_message(message)
        cls.get_sdk().flush()

    @classmethod
    def get_sdk(cls):
        if is_test_env() or is_dev_env():
            return DevSDK

        return sentry_sdk



import random
import string

import abstra.messages as am
import abstra.workflows as aw

# Stage Title: Send Welcome Email
# Retrieve data from the previous stage
name = aw.get_data("name")
email = aw.get_data("email")
department = aw.get_data("department")

# Generate a unique password
password = "".join(random.choices(string.ascii_letters + string.digits, k=12))

# Create a personalized welcome message
welcome_message = f"""
Hi {name},

Welcome to the {department} department! We're excited to have you on board.

Here are your login credentials for the employee portal:
- **Username**: {email}
- **Password**: {password}

You can access the employee portal using the following link:
[Employee Portal](https://employee-portal.example.com)

Please change your password upon first login.

Best regards,
The HR Team
"""

# Send the welcome email
am.send_email(to=str(email), title="Welcome to the Company!", message=welcome_message)

# Store the password in the current thread (optional)
aw.set_data("password", password)



import flask

from abstra_internals.contracts_generated import (
    AbstraLibApiEditorLintersCheckResponse,
    AbstraLibApiEditorLintersFixResponse,
    AbstraLibApiEditorLintersRule,
)
from abstra_internals.linter.rules import rules
from abstra_internals.usage import usage


def check_linters() -> AbstraLibApiEditorLintersCheckResponse:
    checks = []

    for rule in rules:
        try:
            checks.append(AbstraLibApiEditorLintersRule.from_dict(rule.to_dict()))
        except Exception:
            pass

    return checks


def fix_linter(rule_name: str, fix_name: str):
    for rule in rules:
        if rule.name == rule_name:
            for issue in rule.find_issues():
                for fix in issue.fixes:
                    if fix.name == fix_name:
                        fix.fix()
                        return True
    raise Exception(f"Could not find fix {fix_name} for rule {rule_name}")


def fix_all_linters():
    for rule in rules:
        if rule.type != "info":
            for issue in rule.find_issues():
                for fix in issue.fixes:
                    fix.fix()


def get_editor_bp():
    bp = flask.Blueprint("editor_linters", __name__)

    # 1s pooling in this route
    @bp.get("/check")
    def _check_linters():
        return check_linters()

    @bp.post("/fix/<rule_name>/<fix_name>")
    @usage
    def _fix_linter(rule_name: str, fix_name: str):
        fix_linter(rule_name, fix_name)
        return AbstraLibApiEditorLintersFixResponse(success=True).to_dict()

    return bp



from typing import Optional

from abstra_internals.credentials import set_credentials
from abstra_internals.repositories.project.project import ProjectRepository


def start(credential_token: Optional[str] = None):
    if ProjectRepository.exists():
        print("This directory already has an Abstra project.")
        print("\nRun 'abstra editor' in the directory to start the editor.")
        return

    ProjectRepository.generate_getting_started_project()

    if credential_token:
        set_credentials(credential_token)

    print("Project initialized successfully.")
    print("\nRun 'abstra editor' in the directory to start the editor.")



import typing
from typing import Optional, Tuple

import flask
import requests

from abstra_internals.email_templates import authn_template
from abstra_internals.environment import (
    OIDC_AUTHORITY,
)
from abstra_internals.logger import AbstraLogger
from abstra_internals.repositories.email import EmailRepository
from abstra_internals.repositories.jwt_signer import JWTRepository
from abstra_internals.repositories.keyvalue import KVRepository
from abstra_internals.server.controller.main import MainController
from abstra_internals.utils.email import is_valid_email


def authn_kv_key(email: str) -> str:
    return f"authn::key::{email}"


def authn_kv_counter(email: str) -> str:
    return f"authn::counter::{email}"


def get_oidc_userinfo_email(access_token: str) -> typing.Optional[str]:
    if not OIDC_AUTHORITY():
        raise ValueError("[OIDC] OIDC_AUTHORITY is not set")

    r = requests.get(f"{OIDC_AUTHORITY()}/.well-known/openid-configuration")
    if not r.ok:
        raise ValueError("[OIDC] Failed to fetch OIDC configuration")

    userinfo_endpoint = r.json().get("userinfo_endpoint")
    if not userinfo_endpoint:
        raise ValueError("[OIDC] userinfo_endpoint not found in OIDC configuration")

    r = requests.get(
        userinfo_endpoint, headers={"Authorization": f"Bearer {access_token}"}
    )
    if not r.ok:
        return None

    return r.json().get("email")


class AuthController:
    kv_repository: KVRepository
    jwt_repository: JWTRepository
    email_repository: EmailRepository

    def __init__(
        self,
        *,
        kv_repository: KVRepository,
        jwt_repository: JWTRepository,
        email_repository: EmailRepository,
    ):
        self.kv_repository = kv_repository
        self.jwt_repository = jwt_repository
        self.email_repository = email_repository

    def authenticate(self, email: str) -> bool:
        try:
            key = authn_kv_key(email)
            code = self.jwt_repository.gen_code()
            self.kv_repository.set(key, code)

            counter_key = authn_kv_counter(email)
            self.kv_repository.set(counter_key, ".")

            authentication_email = authn_template.generate_email(email, code)
            self.email_repository.send(authentication_email)

            return True
        except Exception as e:
            AbstraLogger.capture_exception(e)
            return False

    def verify(self, email: str, user_code: str) -> Tuple[int, Optional[str]]:
        try:
            key = authn_kv_key(email)
            counter_key = authn_kv_counter(email)
            user_code = self.jwt_repository.sanitize_code(user_code)

            code = self.kv_repository.get(key)
            if code != user_code:
                counter = self.kv_repository.get(counter_key)
                if len(counter) >= 3:
                    self.kv_repository.delete(key)
                    self.kv_repository.delete(counter_key)
                    return 410, None

                self.kv_repository.set(counter_key, counter + ".")
                return 400, None

            self.kv_repository.delete(key)
            self.kv_repository.delete(counter_key)

            return 200, self.jwt_repository.sign(email)
        except Exception as e:
            AbstraLogger.capture_exception(e)
            return 500, None

    def oidc_verify(self, access_token: str) -> Optional[str]:
        try:
            email = get_oidc_userinfo_email(access_token)
            if not email:
                return None

            return self.jwt_repository.sign(email)
        except Exception as e:
            AbstraLogger.capture_exception(e)
            return None


def get_player_bp(main_controller: MainController):
    bp = flask.Blueprint("auth", __name__)
    controller = AuthController(
        kv_repository=main_controller.kv_repository,
        jwt_repository=main_controller.jwt_repository,
        email_repository=main_controller.email_repository,
    )

    @bp.post("/authenticate")
    def _authenticate():
        data = flask.request.get_json(force=True)
        if not data:
            return flask.abort(400)

        email = data.get("email")
        if not email or not is_valid_email(email):
            return flask.abort(400)

        ok = controller.authenticate(email)
        if not ok:
            return flask.abort(400)

        return {"ok": True}

    @bp.post("/verify")
    def _verify():
        data = flask.request.get_json(force=True)
        if not data:
            return flask.abort(400)

        email = data.get("email")
        code = data.get("token")
        if not email or not code:
            return flask.abort(400)

        status, jwt = controller.verify(email, code)
        if not jwt:
            return flask.abort(status)

        return {"jwt": jwt}

    @bp.post("/oidc-verify")
    def _oidc_access():
        data = flask.request.get_json(force=True)
        if not data:
            return flask.abort(400)

        access_token = data.get("access_token")
        if not access_token:
            return flask.abort(400)

        jwt = controller.oidc_verify(access_token)
        if not jwt:
            return flask.abort(404)

        return {"jwt": jwt}

    return bp



import threading
from dataclasses import dataclass
from typing import List

from abstra_internals.controllers.execution_client import ExecutionClient
from abstra_internals.controllers.execution_client_form import FormClient
from abstra_internals.controllers.execution_client_hook import HookClient
from abstra_internals.entities.execution import Execution


class ExecutionStoreException(Exception):
    pass


class ExecutionNotFound(ExecutionStoreException):
    pass


class ClientTypeMismatch(ExecutionStoreException):
    pass


@dataclass
class ExecutionStoreData:
    client: ExecutionClient
    execution: Execution
    thread_id: int


class ExecutionStore:
    rlock = threading.RLock()
    store: List[ExecutionStoreData] = []

    @classmethod
    def clear(cls):
        data = cls.get_by_thread()
        with cls.rlock:
            cls.store.remove(data)

    @classmethod
    def set(cls, execution: Execution, client: ExecutionClient):
        with cls.rlock:
            cls.store.append(
                ExecutionStoreData(
                    thread_id=threading.get_ident(),
                    execution=execution,
                    client=client,
                )
            )

    @classmethod
    def get_by_thread(cls) -> ExecutionStoreData:
        thread_id = threading.get_ident()
        with cls.rlock:
            for data in cls.store:
                if data.thread_id == thread_id:
                    return data
            raise ExecutionNotFound()

    @classmethod
    def get_by_execution_id(cls, execution_id: str) -> ExecutionStoreData:
        with cls.rlock:
            for data in cls.store:
                if data.execution.id == execution_id:
                    return data
            raise ExecutionNotFound()

    @classmethod
    def get_hook_client(cls) -> HookClient:
        client = cls.get_by_thread().client
        if not isinstance(client, HookClient):
            raise ClientTypeMismatch()
        return client

    @classmethod
    def get_form_client(cls) -> FormClient:
        client = cls.get_by_thread().client
        if not isinstance(client, FormClient):
            raise ClientTypeMismatch()
        return client



import flask
import flask_cors

from abstra_internals.server.blueprints.player import get_player_bp
from abstra_internals.server.controller.main import MainController


def get_cloud_app(controller: MainController):
    app = flask.Flask(__name__)
    app.config["SOCK_SERVER_OPTIONS"] = {"subprotocols": ["default"]}
    app.url_map.strict_slashes = False
    flask_cors.CORS(app)

    player = get_player_bp(controller)
    app.register_blueprint(player)

    return app



from unittest import TestCase

from abstra_internals.repositories.project.project import (
    FormStage,
    JobStage,
    NotificationTrigger,
    ProjectRepository,
    ScriptStage,
    WorkflowTransition,
)
from abstra_internals.server.controller.workflows import (
    get_path,
    get_workflow,
    update_workflow,
)
from tests.fixtures import clear_dir, init_dir


class WorkflowTest(TestCase):
    def setUp(self):
        self.root = init_dir()

    def tearDown(self) -> None:
        clear_dir(self.root)

    def test_empty_project(self):
        path = get_path(3)
        self.assertEqual(path, [])

    def test_get_empty(self):
        workflow = get_workflow()
        self.assertEqual(workflow, {"stages": [], "transitions": []})

    def test_get_with_stages(self):
        initial_state = {
            "stages": [
                {
                    "type": "forms",
                    "id": "foo",
                    "title": "Foo",
                    "props": {
                        "filename": "foo.py",
                        "path": "foo",
                        "variableName": None,
                        "itemName": None,
                    },
                    "position": {"x": 0, "y": 0},
                }
            ],
            "transitions": [],
        }
        update_workflow(initial_state)
        workflow = get_workflow()
        self.assertEqual(workflow, initial_state)

    def test_get_with_transitions(self):
        initial_state = {
            "stages": [
                {
                    "type": "forms",
                    "id": "foo",
                    "title": "Foo",
                    "props": {
                        "filename": "foo.py",
                        "path": "foo",
                        "variableName": None,
                        "itemName": None,
                    },
                    "position": {"x": 0, "y": 0},
                },
                {
                    "type": "hooks",
                    "id": "bar",
                    "props": {
                        "filename": "bar.py",
                        "path": "bar",
                        "variableName": None,
                        "itemName": None,
                    },
                    "title": "Bar",
                    "position": {"x": 0, "y": 0},
                },
            ],
            "transitions": [
                {
                    "id": "foo",
                    "sourceStageId": "foo",
                    "targetStageId": "bar",
                    "type": "forms:finished",
                    "props": {"conditionValue": None},
                }
            ],
        }

        updated_workflow = update_workflow(initial_state)

        self.assertEqual(updated_workflow, initial_state)

        project = ProjectRepository.load()

        self.assertEqual(len(project.forms[0].workflow_transitions), 1)

        gotten_workflow = get_workflow()
        self.assertEqual(gotten_workflow, initial_state)

        initial_state = {
            "stages": [
                {
                    "type": "forms",
                    "id": "foo",
                    "props": {
                        "filename": "foo.py",
                        "path": "foo",
                        "variableName": None,
                        "itemName": None,
                    },
                    "title": "Foo",
                    "position": {"x": 0, "y": 0},
                },
                {
                    "type": "hooks",
                    "id": "bar",
                    "props": {
                        "filename": "bar.py",
                        "path": "bar",
                        "variableName": None,
                        "itemName": None,
                    },
                    "title": "Bar",
                    "position": {"x": 0, "y": 0},
                },
            ],
            "transitions": [
                {
                    "id": "foo",
                    "sourceStageId": "foo",
                    "targetStageId": "bar",
                    "type": "foo",
                    "props": {"conditionValue": None},
                }
            ],
        }

        updated_workflow = update_workflow(initial_state)

        self.assertEqual(updated_workflow, initial_state)

    def test_update_add_stages(self):
        update_workflow(
            {
                "stages": [
                    {
                        "type": "forms",
                        "id": "foo",
                        "title": "Foo",
                        "position": {"x": 0, "y": 0},
                        "props": {
                            "variableName": None,
                            "path": "foo",
                            "filename": "foo.py",
                            "itemName": None,
                        },
                    }
                ],
                "transitions": [],
            }
        )
        workflow = get_workflow()
        self.assertEqual(len(workflow["stages"]), 1)

    def test_update_remove_stages(self):
        update_workflow(
            {
                "stages": [
                    {
                        "type": "forms",
                        "id": "foo",
                        "title": "Foo",
                        "position": {"x": 0, "y": 0},
                        "props": {
                            "variableName": None,
                            "path": "foo",
                            "filename": "foo.py",
                            "itemName": None,
                        },
                    }
                ],
                "transitions": [],
            }
        )
        workflow = get_workflow()
        self.assertEqual(len(workflow["stages"]), 1)
        update_workflow(
            {
                "stages": [],
                "transitions": [],
            }
        )
        workflow = get_workflow()
        self.assertEqual(len(workflow["stages"]), 0)

    def test_update_change_stage_position(self):
        update_workflow(
            {
                "stages": [
                    {
                        "type": "forms",
                        "id": "foo",
                        "title": "Foo",
                        "position": {"x": 0, "y": 0},
                        "props": {
                            "variableName": None,
                            "path": "foo",
                            "filename": "foo.py",
                            "itemName": None,
                        },
                    }
                ],
                "transitions": [],
            }
        )
        workflow = get_workflow()
        self.assertEqual(len(workflow["stages"]), 1)
        self.assertEqual(workflow["stages"][0]["position"], {"x": 0, "y": 0})
        update_workflow(
            {
                "stages": [
                    {
                        "type": "forms",
                        "id": "foo",
                        "title": "Foo",
                        "position": {"x": 1, "y": 1},
                        "props": {
                            "filename": "foo.py",
                            "path": "foo",
                            "variableName": None,
                            "itemName": None,
                        },
                    }
                ],
                "transitions": [],
            }
        )
        workflow = get_workflow()
        self.assertEqual(len(workflow["stages"]), 1)
        self.assertEqual(workflow["stages"][0]["position"], {"x": 1, "y": 1})

    def test_update_change_stage_title(self):
        update_workflow(
            {
                "stages": [
                    {
                        "type": "forms",
                        "id": "foo",
                        "title": "Foo",
                        "position": {"x": 0, "y": 0},
                        "props": {
                            "variableName": None,
                            "path": "foo",
                            "filename": "foo.py",
                            "itemName": None,
                        },
                    }
                ],
                "transitions": [],
            }
        )
        workflow = get_workflow()
        self.assertEqual(len(workflow["stages"]), 1)
        self.assertEqual(workflow["stages"][0]["title"], "Foo")
        update_workflow(
            {
                "stages": [
                    {
                        "type": "forms",
                        "id": "foo",
                        "title": "Bar",
                        "position": {"x": 0, "y": 0},
                        "props": {
                            "filename": "bar.py",
                            "path": "bar",
                            "variableName": None,
                            "itemName": None,
                        },
                    }
                ],
                "transitions": [],
            }
        )
        workflow = get_workflow()
        self.assertEqual(len(workflow["stages"]), 1)
        self.assertEqual(workflow["stages"][0]["title"], "Bar")

    def test_update_add_transitions(self):
        update_workflow(
            {
                "stages": [
                    {
                        "type": "forms",
                        "id": "foo",
                        "title": "Foo",
                        "position": {"x": 0, "y": 0},
                        "props": {
                            "variableName": None,
                            "path": "foo",
                            "filename": "foo.py",
                            "itemName": None,
                        },
                    },
                    {
                        "type": "hooks",
                        "id": "bar",
                        "title": "Bar",
                        "position": {"x": 0, "y": 0},
                        "props": {
                            "filename": "bar.py",
                            "path": "bar",
                            "variableName": None,
                            "itemName": None,
                        },
                    },
                ],
                "transitions": [],
            }
        )
        workflow = get_workflow()
        self.assertEqual(len(workflow["transitions"]), 0)
        update_workflow(
            {
                "stages": [
                    {
                        "type": "forms",
                        "id": "foo",
                        "title": "Foo",
                        "position": {"x": 0, "y": 0},
                        "props": {
                            "variableName": None,
                            "path": "foo",
                            "filename": "foo.py",
                            "itemName": None,
                        },
                    },
                    {
                        "type": "hooks",
                        "id": "bar",
                        "title": "Bar",
                        "position": {"x": 0, "y": 0},
                        "props": {
                            "filename": "bar.py",
                            "path": "bar",
                            "variableName": None,
                            "itemName": None,
                        },
                    },
                ],
                "transitions": [
                    {
                        "id": "foo",
                        "sourceStageId": "foo",
                        "targetStageId": "bar",
                        "type": "foo",
                        "conditionValue": None,
                    }
                ],
            }
        )
        workflow = get_workflow()
        self.assertEqual(len(workflow["transitions"]), 1)

    def test_update_remove_transitions(self):
        update_workflow(
            {
                "stages": [
                    {
                        "type": "forms",
                        "id": "foo",
                        "title": "Foo",
                        "position": {"x": 0, "y": 0},
                        "props": {
                            "variableName": None,
                            "path": "foo",
                            "filename": "foo.py",
                            "itemName": None,
                        },
                    },
                    {
                        "type": "hooks",
                        "id": "bar",
                        "title": "Bar",
                        "position": {"x": 0, "y": 0},
                        "props": {
                            "filename": "bar.py",
                            "path": "bar",
                            "variableName": None,
                            "itemName": None,
                        },
                    },
                ],
                "transitions": [
                    {
                        "id": "foo",
                        "sourceStageId": "foo",
                        "targetStageId": "bar",
                        "type": "foo",
                        "conditionValue": None,
                    }
                ],
            }
        )
        workflow = get_workflow()
        self.assertEqual(len(workflow["transitions"]), 1)
        update_workflow(
            {
                "stages": [],
                "transitions": [],
            }
        )
        workflow = get_workflow()
        self.assertEqual(len(workflow["transitions"]), 0)

    def test_update_change_transition_type(self):
        update_workflow(
            {
                "stages": [
                    {
                        "type": "forms",
                        "id": "foo",
                        "title": "Foo",
                        "position": {"x": 0, "y": 0},
                        "props": {
                            "variableName": None,
                            "path": "foo",
                            "filename": "foo.py",
                            "itemName": None,
                        },
                    },
                    {
                        "type": "hooks",
                        "id": "bar",
                        "title": "Bar",
                        "position": {"x": 0, "y": 0},
                        "props": {
                            "filename": "bar.py",
                            "path": "bar",
                            "variableName": None,
                            "itemName": None,
                        },
                    },
                ],
                "transitions": [
                    {
                        "id": "foo",
                        "sourceStageId": "foo",
                        "targetStageId": "bar",
                        "type": "foo",
                        "conditionValue": None,
                    }
                ],
            }
        )
        workflow = get_workflow()
        self.assertEqual(len(workflow["transitions"]), 1)
        self.assertEqual(workflow["transitions"][0]["type"], "foo")
        update_workflow(
            {
                "stages": [
                    {
                        "type": "forms",
                        "id": "foo",
                        "title": "Foo",
                        "position": {"x": 0, "y": 0},
                        "props": {
                            "variableName": None,
                            "path": "foo",
                            "filename": "foo.py",
                            "itemName": None,
                        },
                    },
                    {
                        "type": "hooks",
                        "id": "bar",
                        "title": "Bar",
                        "position": {"x": 0, "y": 0},
                        "props": {
                            "filename": "bar.py",
                            "path": "bar",
                            "variableName": None,
                            "itemName": None,
                        },
                    },
                ],
                "transitions": [
                    {
                        "id": "foo",
                        "sourceStageId": "foo",
                        "targetStageId": "bar",
                        "type": "bar",
                    }
                ],
            }
        )
        workflow = get_workflow()
        self.assertEqual(len(workflow["transitions"]), 1)
        self.assertEqual(workflow["transitions"][0]["type"], "bar")

    def test_single_form(self):
        project = ProjectRepository.load()
        form = FormStage(
            id="form",
            path="form",
            title="Form",
            file="form.py",
            workflow_position=(0, 0),
            workflow_transitions=[],
            notification_trigger=NotificationTrigger(
                variable_name="val", enabled=False
            ),
        )
        project.forms.append(form)
        ProjectRepository.save(project)

        path = get_path(3)
        self.assertEqual(path, ["form"])

    def test_more_then_one_stage(self):
        project = ProjectRepository.load()
        form = FormStage(
            id="form",
            path="form",
            title="Form",
            file="form.py",
            workflow_position=(0, 0),
            workflow_transitions=[
                WorkflowTransition(
                    id="1",
                    target_type="script",
                    target_id="script",
                    type="finished",
                )
            ],
            notification_trigger=NotificationTrigger(
                variable_name="val", enabled=False
            ),
        )
        script = ScriptStage(
            id="script",
            file="script.py",
            title="Script",
            workflow_position=(0, 0),
            workflow_transitions=[],
        )
        project.forms.append(form)
        project.scripts.append(script)
        ProjectRepository.save(project)

        path = get_path(2)
        self.assertEqual(path, ["form", "script"])

    def test_more_stages_then_limit(self):
        project = ProjectRepository.load()
        form = FormStage(
            id="form",
            path="form",
            title="Form",
            file="form.py",
            workflow_position=(0, 0),
            workflow_transitions=[
                WorkflowTransition(
                    id="1",
                    target_type="script",
                    target_id="script",
                    type="finished",
                )
            ],
            notification_trigger=NotificationTrigger(
                variable_name="val", enabled=False
            ),
        )
        script = ScriptStage(
            id="script",
            file="script.py",
            title="Script",
            workflow_position=(0, 0),
            workflow_transitions=[],
        )
        project.forms.append(form)
        project.scripts.append(script)
        ProjectRepository.save(project)

        path = get_path(1)
        self.assertEqual(path, ["form"])

    def test_more_then_one_path(self):
        project = ProjectRepository.load()
        form0 = FormStage(
            id="form0",
            path="form0",
            title="Form0",
            file="form0.py",
            workflow_position=(0, 0),
            workflow_transitions=[
                WorkflowTransition(
                    id="1",
                    target_type="script",
                    target_id="script1",
                    type="finished",
                ),
                WorkflowTransition(
                    id="2",
                    target_type="script",
                    target_id="script2",
                    type="finished",
                ),
            ],
            notification_trigger=NotificationTrigger(
                variable_name="val", enabled=False
            ),
        )
        script1 = ScriptStage(
            id="script1",
            file="script1.py",
            title="Script1",
            workflow_position=(0, 0),
            workflow_transitions=[],
        )

        script2 = ScriptStage(
            id="script2",
            file="script2.py",
            title="Script2",
            workflow_position=(0, 0),
            workflow_transitions=[
                WorkflowTransition(
                    id="3",
                    target_type="script",
                    target_id="script3",
                    type="finished",
                )
            ],
        )

        script3 = ScriptStage(
            id="script3",
            file="script3.py",
            title="Script3",
            workflow_position=(0, 0),
            workflow_transitions=[],
        )

        project.forms.append(form0)
        project.scripts.append(script1)
        project.scripts.append(script2)
        project.scripts.append(script3)
        ProjectRepository.save(project)

        path = get_path(3)
        self.assertEqual(path, ["form0", "script2", "script3"])

    def test_more_than_one_initial_stage(self):
        project = ProjectRepository.load()
        form0 = FormStage(
            id="form0",
            path="form0",
            title="Form0",
            file="form0.py",
            workflow_position=(0, 0),
            workflow_transitions=[
                WorkflowTransition(
                    id="1",
                    target_type="script",
                    target_id="script1",
                    type="finished",
                ),
                WorkflowTransition(
                    id="2",
                    target_type="script",
                    target_id="script2",
                    type="finished",
                ),
            ],
            notification_trigger=NotificationTrigger(
                variable_name="val", enabled=False
            ),
        )
        script1 = ScriptStage(
            id="script1",
            file="script1.py",
            title="Script1",
            workflow_position=(0, 0),
            workflow_transitions=[],
        )

        script2 = ScriptStage(
            id="script2",
            file="script2.py",
            title="Script2",
            workflow_position=(0, 0),
            workflow_transitions=[],
        )

        form1 = FormStage(
            id="form1",
            path="form1",
            title="Form1",
            file="form1.py",
            workflow_position=(0, 0),
            workflow_transitions=[
                WorkflowTransition(
                    id="3",
                    target_type="form",
                    target_id="form2",
                    type="finished",
                ),
            ],
            notification_trigger=NotificationTrigger(
                variable_name="val", enabled=False
            ),
        )

        form2 = FormStage(
            id="form2",
            path="form2",
            title="Form2",
            file="form2.py",
            workflow_position=(0, 0),
            workflow_transitions=[
                WorkflowTransition(
                    id="4",
                    target_type="script",
                    target_id="script2",
                    type="finished",
                ),
            ],
            notification_trigger=NotificationTrigger(
                variable_name="val", enabled=False
            ),
        )

        project.forms.append(form0)
        project.scripts.append(script1)
        project.scripts.append(script2)
        project.forms.append(form1)
        project.forms.append(form2)
        ProjectRepository.save(project)

        path = get_path(3)
        self.assertEqual(path, ["form1", "form2", "script2"])

    def test_disjointed_graphs(self):
        project = ProjectRepository.load()
        form0 = FormStage(
            id="form0",
            path="form0",
            title="Form0",
            file="form0.py",
            workflow_position=(0, 0),
            workflow_transitions=[
                WorkflowTransition(
                    id="1",
                    target_type="script",
                    target_id="script1",
                    type="finished",
                ),
                WorkflowTransition(
                    id="2",
                    target_type="script",
                    target_id="script2",
                    type="finished",
                ),
            ],
            notification_trigger=NotificationTrigger(
                variable_name="val", enabled=False
            ),
        )
        script1 = ScriptStage(
            id="script1",
            file="script1.py",
            title="Script1",
            workflow_position=(0, 0),
            workflow_transitions=[],
        )

        script2 = ScriptStage(
            id="script2",
            file="script2.py",
            title="Script2",
            workflow_position=(0, 0),
            workflow_transitions=[],
        )

        form1 = FormStage(
            id="form1",
            path="form1",
            title="Form1",
            file="form1.py",
            workflow_position=(0, 0),
            workflow_transitions=[
                WorkflowTransition(
                    id="3",
                    target_type="form",
                    target_id="form2",
                    type="finished",
                ),
            ],
            notification_trigger=NotificationTrigger(
                variable_name="val", enabled=False
            ),
        )

        form2 = FormStage(
            id="form2",
            path="form2",
            title="Form2",
            file="form2.py",
            workflow_position=(0, 0),
            workflow_transitions=[
                WorkflowTransition(
                    id="4",
                    target_type="form",
                    target_id="form3",
                    type="finished",
                ),
            ],
            notification_trigger=NotificationTrigger(
                variable_name="val", enabled=False
            ),
        )

        form3 = FormStage(
            id="form3",
            path="form3",
            title="Form3",
            file="form3.py",
            workflow_position=(0, 0),
            workflow_transitions=[],
            notification_trigger=NotificationTrigger(
                variable_name="val", enabled=False
            ),
        )

        project.forms.append(form0)
        project.scripts.append(script1)
        project.scripts.append(script2)
        project.forms.append(form1)
        project.forms.append(form2)
        project.forms.append(form3)
        ProjectRepository.save(project)

        path = get_path(3)
        self.assertEqual(path, ["form1", "form2", "form3"])

    def test_simple_loop(self):
        project = ProjectRepository.load()
        job0 = JobStage(
            id="job0",
            title="Job0",
            file="job0.py",
            workflow_position=(0, 0),
            schedule="* * * * *",
            workflow_transitions=[
                WorkflowTransition(
                    id="0",
                    target_type="form",
                    target_id="form0",
                    type="finished",
                ),
            ],
        )
        form0 = FormStage(
            id="form0",
            path="form0",
            title="Form0",
            file="form0.py",
            workflow_position=(0, 0),
            workflow_transitions=[
                WorkflowTransition(
                    id="1",
                    target_type="script",
                    target_id="script1",
                    type="finished",
                ),
            ],
            notification_trigger=NotificationTrigger(
                variable_name="val", enabled=False
            ),
        )
        script1 = ScriptStage(
            id="script1",
            file="script1.py",
            title="Script1",
            workflow_position=(0, 0),
            workflow_transitions=[
                WorkflowTransition(
                    id="2",
                    target_type="form",
                    target_id="form0",
                    type="finished",
                ),
            ],
        )

        project.jobs.append(job0)
        project.forms.append(form0)
        project.scripts.append(script1)
        ProjectRepository.save(project)

        path = get_path(3)
        self.assertEqual(path, ["job0", "form0", "script1"])

    def test_dont_enter_in_loops(self):
        project = ProjectRepository.load()
        job0 = JobStage(
            id="job0",
            title="Job0",
            file="job0.py",
            workflow_position=(0, 0),
            schedule="* * * * *",
            workflow_transitions=[
                WorkflowTransition(
                    id="0",
                    target_type="form",
                    target_id="form0",
                    type="finished",
                ),
            ],
        )
        form0 = FormStage(
            id="form0",
            path="form0",
            title="Form0",
            file="form0.py",
            workflow_position=(0, 0),
            workflow_transitions=[
                WorkflowTransition(
                    id="1",
                    target_type="script",
                    target_id="script1",
                    type="finished",
                ),
            ],
            notification_trigger=NotificationTrigger(
                variable_name="val", enabled=False
            ),
        )
        script1 = ScriptStage(
            id="script1",
            file="script1.py",
            title="Script1",
            workflow_position=(0, 0),
            workflow_transitions=[
                WorkflowTransition(
                    id="2",
                    target_type="form",
                    target_id="form0",
                    type="finished",
                ),
            ],
        )

        project.jobs.append(job0)
        project.forms.append(form0)
        project.scripts.append(script1)
        ProjectRepository.save(project)

        path = get_path(10)
        self.assertEqual(path, ["job0", "form0", "script1"])

    def test_clips_the_largest_path(self):
        project = ProjectRepository.load()
        form0 = FormStage(
            id="form0",
            path="form0",
            title="Form0",
            file="form0.py",
            workflow_position=(0, 0),
            workflow_transitions=[
                WorkflowTransition(
                    id="1",
                    target_type="form",
                    target_id="form1",
                    type="finished",
                ),
            ],
            notification_trigger=NotificationTrigger(
                variable_name="val", enabled=False
            ),
        )

        form1 = FormStage(
            id="form1",
            path="form1",
            title="Form1",
            file="form1.py",
            workflow_position=(0, 0),
            workflow_transitions=[
                WorkflowTransition(
                    id="2",
                    target_type="form",
                    target_id="form2",
                    type="finished",
                ),
                WorkflowTransition(
                    id="3",
                    target_type="script",
                    target_id="script1",
                    type="finished",
                ),
            ],
            notification_trigger=NotificationTrigger(
                variable_name="val", enabled=False
            ),
        )

        script1 = ScriptStage(
            id="script1",
            file="script1.py",
            title="Script1",
            workflow_position=(0, 0),
            workflow_transitions=[],
        )

        form2 = FormStage(
            id="form2",
            path="form2",
            title="Form2",
            file="form2.py",
            workflow_position=(0, 0),
            workflow_transitions=[
                WorkflowTransition(
                    id="4",
                    target_type="form",
                    target_id="form3",
                    type="finished",
                ),
                WorkflowTransition(
                    id="5",
                    target_type="script",
                    target_id="script2",
                    type="finished",
                ),
            ],
            notification_trigger=NotificationTrigger(
                variable_name="val", enabled=False
            ),
        )

        script2 = ScriptStage(
            id="script2",
            file="script2.py",
            title="Script2",
            workflow_position=(0, 0),
            workflow_transitions=[],
        )

        form3 = FormStage(
            id="form3",
            path="form3",
            title="Form3",
            file="form3.py",
            workflow_position=(0, 0),
            workflow_transitions=[
                WorkflowTransition(
                    id="7",
                    target_type="script",
                    target_id="script3",
                    type="finished",
                ),
            ],
            notification_trigger=NotificationTrigger(
                variable_name="val", enabled=False
            ),
        )

        script3 = ScriptStage(
            id="script3",
            file="script3.py",
            title="Script3",
            workflow_position=(0, 0),
            workflow_transitions=[],
        )

        project.forms.append(form0)
        project.forms.append(form1)
        project.scripts.append(script1)
        project.forms.append(form2)
        project.scripts.append(script2)
        project.forms.append(form3)
        project.scripts.append(script3)
        ProjectRepository.save(project)

        path = get_path(3)
        self.assertEqual(path, ["form0", "form1", "form2"])



from typing import Optional

import requests

from abstra_internals.contracts_generated import (
    CloudApiCliApiKeyInfoResponse,
    CloudApiCliBuildCreateResponse,
)
from abstra_internals.environment import CLOUD_API_ENDPOINT
from abstra_internals.logger import AbstraLogger


def create_build(headers: dict) -> CloudApiCliBuildCreateResponse:
    url = f"{CLOUD_API_ENDPOINT}/cli/builds"
    r = requests.post(url=url, headers=headers)
    if not r.ok:
        raise Exception(f"Failed to create build: {r.text}")

    data = r.json()
    return CloudApiCliBuildCreateResponse.from_dict(data)


def update_build(build_id: str, headers: dict, error: Optional[str] = None):
    url = f"{CLOUD_API_ENDPOINT}/cli/builds/{build_id}"
    requests.patch(url=url, headers=headers, json=dict(error=error)).raise_for_status()


def get_api_key_info(headers: dict) -> dict:
    url = f"{CLOUD_API_ENDPOINT}/cli/api-keys/info"
    try:
        response = requests.get(url, headers=headers)
    except Exception as e:
        AbstraLogger.capture_exception(e)
        return {"logged": False, "reason": "CONNECTION_ERROR"}

    if response.ok:
        response_data = CloudApiCliApiKeyInfoResponse.from_dict(response.json())
        return {"logged": True, "info": response_data.to_dict()}
    else:
        return {"logged": False, "reason": "INVALID_API_TOKEN"}


def get_ai_messages(messages, stage, thread_id, headers: dict):
    url = f"{CLOUD_API_ENDPOINT}/cli/ai/messages"
    body = {"messages": messages, "runtime": stage, "threadId": thread_id}
    return requests.post(url, headers=headers, json=body, stream=True).iter_content(
        chunk_size=None
    )


def generate_project(prompt: str, abstra_json_version: str, headers: dict):
    url = f"{CLOUD_API_ENDPOINT}/cli/ai/generate"
    body = {"prompt": prompt, "version": abstra_json_version}
    r = requests.post(url, headers=headers, json=body)
    r.raise_for_status()
    return r.json()


def create_thread(headers: dict):
    url = f"{CLOUD_API_ENDPOINT}/cli/ai/thread"
    r = requests.post(url, headers=headers)
    r.raise_for_status()
    return r.json()


def cancel_all(headers: dict, thread_id: str):
    url = f"{CLOUD_API_ENDPOINT}/cli/ai/cancel-all"
    r = requests.post(url, headers=headers, json={"threadId": thread_id})
    r.raise_for_status()
    return r.json()


def get_project_info(headers):
    url = f"{CLOUD_API_ENDPOINT}/cli/project"
    r = requests.get(url, headers=headers)
    r.raise_for_status()
    return r.json()



import abc


class ExecutionClient(abc.ABC):
    @abc.abstractmethod
    def handle_lock_failed(self, status: str) -> None:
        raise NotImplementedError()

    @abc.abstractmethod
    def handle_failure(self, e: Exception) -> None:
        raise NotImplementedError()

    @abc.abstractmethod
    def handle_success(self) -> None:
        raise NotImplementedError()

    @abc.abstractmethod
    def handle_start(self, execution_id: str):
        raise NotImplementedError()

    @abc.abstractmethod
    def handle_unset_thread(self):
        raise NotImplementedError()


# TODO: inherit all clients from BasicClient - same constructor
class BasicClient(ExecutionClient):
    def handle_lock_failed(self, status: str) -> None:
        pass

    def handle_failure(self, e: Exception) -> None:
        pass

    def handle_success(self) -> None:
        pass

    def handle_start(self, execution_id: str):
        pass

    def handle_unset_thread(self):
        pass



import os

WORKERS = os.getenv("ABSTRA_WORKERS", 2)
THREADS = os.getenv("ABSTRA_THREADS", 20)
WORKER_TEMP_DIR = os.getenv("ABSTRA_WORKER_TEMP_DIR")
WORKER_CLASS = os.getenv("ABSTRA_WORKER_CLASS", "gthread")

HOST = os.getenv("ABSTRA_HOST", "localhost")
DEFAULT_PORT = os.getenv("PORT") or os.getenv("ABSTRA_SERVER_PORT")

BUILD_ID = os.getenv("ABSTRA_BUILD_ID") or "dev"
PROJECT_ID = os.getenv("ABSTRA_PROJECT_ID") or "dev-project-id"
PROJECT_URL = os.getenv("ABSTRA_PROJECT_URL")

IS_PRODUCTION = os.getenv("ABSTRA_ENVIRONMENT") == "production"
SHOW_WATERMARK = os.getenv("ABSTRA_SHOW_WATERMARK", "false") == "true"

CLOUD_API_ENDPOINT = os.getenv("CLOUD_API_ENDPOINT") or "https://cloud-api.abstra.cloud"
CLOUD_API_CLI_URL = f"{CLOUD_API_ENDPOINT}/cli"

PUBLIC_KEY = os.getenv("ABSTRA_JWT_PUBLIC_KEY_PEM")
FILES_FOLDER = os.getenv("ABSTRA_FILES_FOLDER")

SIDECAR_SHARED_TOKEN = os.getenv("ABSTRA_SIDECAR_SHARED_TOKEN", "shared")
SIDECAR_HEADERS = {"shared-token": SIDECAR_SHARED_TOKEN}
SIDECAR_URL = os.getenv("ABSTRA_SIDECAR_URL")

__WORKER_UUID_ENV__ = "ABSTRA_WORKER_UUID"


def WORKER_UUID():
    return os.getenv(__WORKER_UUID_ENV__)


def set_WORKER_UUID(worker_uuid: str):
    os.environ[__WORKER_UUID_ENV__] = worker_uuid


__SERVER_UUID_ENV__ = "ABSTRA_SERVER_UUID"


def SERVER_UUID():
    return os.getenv(__SERVER_UUID_ENV__)


def set_SERVER_UUID(server_uuid: str):
    os.environ[__SERVER_UUID_ENV__] = server_uuid


def OIDC_CLIENT_ID():
    return os.getenv("ABSTRA_OIDC_CLIENT_ID")


def OIDC_AUTHORITY():
    return os.getenv("ABSTRA_OIDC_AUTHORITY")



from unittest import TestCase

from .api import normalize_format


class TestNormalizeFormat(TestCase):
    def test_normal(self):
        self.assertEqual(
            normalize_format({"a": {"type": "string"}}), {"a": {"type": "string"}}
        )

    def test_string(self):
        self.assertEqual(normalize_format({"a": "string"}), {"a": {"type": "string"}})

    def test_str(self):
        self.assertEqual(normalize_format({"a": str}), {"a": {"type": "string"}})

    def test_int(self):
        self.assertEqual(normalize_format({"a": int}), {"a": {"type": "integer"}})

    def test_float(self):
        self.assertEqual(normalize_format({"a": float}), {"a": {"type": "number"}})

    def test_bool(self):
        self.assertEqual(normalize_format({"a": bool}), {"a": {"type": "boolean"}})

    def test_enum(self):
        self.assertEqual(
            normalize_format({"a": ["b", "c"]}), {"a": {"enum": ["b", "c"]}}
        )



import flask

from abstra_internals.jwt_auth import USER_AUTH_HEADER_KEY, UserClaims
from abstra_internals.repositories.users import UsersRepository
from abstra_internals.server.controller.main import MainController


class UserController:
    def __init__(
        self,
        users_repository: UsersRepository,
    ) -> None:
        self.users_repository = users_repository

    def sign_up(self, claims: UserClaims) -> bool:
        return self.users_repository.insert_user(claims.email)


def get_player_bp(main_controller: MainController):
    controller = UserController(main_controller.users_repository)
    bp = flask.Blueprint("playerUsers", __name__)

    @bp.post("/sign-up")
    def _signup_user():
        auth = flask.request.headers.get(USER_AUTH_HEADER_KEY)
        if auth is None:
            flask.abort(401)

        claims = UserClaims.from_jwt(auth.split(" ")[1])
        if claims is None or claims.email is None:
            flask.abort(401)

        return {"success": controller.sign_up(claims)}

    return bp



import flask

from abstra_internals.server.controller.main import MainController
from abstra_internals.usage import usage


def get_editor_bp(controller: MainController):
    bp = flask.Blueprint("editor_login", __name__)

    @bp.get("/")
    @usage
    def _get_login():
        return controller.get_login()

    @bp.post("/")
    @usage
    def _create_login():
        data = flask.request.json
        if not data:
            flask.abort(400)

        return controller.create_login(data["token"])

    @bp.delete("/")
    @usage
    def _delete_login():
        return controller.delete_login()

    @bp.get("/info")
    def _get_project_info():
        return controller.get_project_info()

    return bp



import abc
import datetime

import jwt
import requests

from abstra_internals.environment import (
    PROJECT_ID,
    SIDECAR_HEADERS,
    SIDECAR_URL,
)
from abstra_internals.utils import generate_n_digit_code


class JWTRepository(abc.ABC):
    def sign(self, email: str):
        raise NotImplementedError()

    def gen_code(self):
        raise NotImplementedError()

    def sanitize_code(self, code: str) -> str:
        raise NotImplementedError()


class ProductionJWTRepository(JWTRepository):
    base_url: str

    def __init__(self, sidecar_url: str):
        self.base_url = sidecar_url + "/jwt/sign"

    def sign(self, email: str):
        r = requests.post(
            self.base_url, json=dict(email=email), headers=SIDECAR_HEADERS
        )

        if not r.ok:
            raise ValueError(f"Failed to sign JWT: {r.text}")

        return r.json()["jwt"]

    def gen_code(self):
        return generate_n_digit_code(6)

    def sanitize_code(self, code: str) -> str:
        return code


class LocalJWTRepository(JWTRepository):
    def sign(self, email: str):
        return jwt.encode(
            key="fake",
            algorithm="HS256",
            payload={
                "email": email,
                "aud": PROJECT_ID,
                "exp": datetime.datetime.utcnow() + datetime.timedelta(days=7),
            },
        )

    def gen_code(self):
        return "000000"

    def sanitize_code(self, code: str) -> str:
        return "000000"


def jwt_repository_factory() -> JWTRepository:
    if SIDECAR_URL is None:
        return LocalJWTRepository()
    else:
        return ProductionJWTRepository(SIDECAR_URL)



import abc
import datetime as dt

import requests

from abstra_internals.environment import SIDECAR_HEADERS, SIDECAR_URL


class LockRepository(abc.ABC):
    def lock(self, key: str, exp: int) -> bool:
        raise NotImplementedError()


class ProductionLockRepository(LockRepository):
    base_url: str

    def __init__(self, sidecar_url: str):
        self.base_url = sidecar_url + "/lock"

    def lock(self, key: str, exp: int) -> bool:
        data = {"key": key, "exp": exp, "unlock": False}
        response = requests.post(self.base_url, json=data, headers=SIDECAR_HEADERS)
        return response.status_code == 200


class LocalLockRepository(LockRepository):
    locks = {}

    def lock(self, key: str, exp: int) -> bool:
        lock = self.locks.get(key)
        can_lock = lock is None or lock < dt.datetime.now()

        if can_lock:
            self.locks[key] = dt.datetime.now() + dt.timedelta(seconds=exp)

        return can_lock


def lock_repository_factory() -> LockRepository:
    if SIDECAR_URL is None:
        return LocalLockRepository()
    else:
        return ProductionLockRepository(SIDECAR_URL)



import flask

from abstra_internals.constants import STAGE_RUN_ID_PARAM_KEY
from abstra_internals.controllers.execution import (
    DetachedExecutionController,
    ExecutionController,
)
from abstra_internals.controllers.execution_client_hook import HookClient
from abstra_internals.entities.execution import context_from_flask
from abstra_internals.server.controller.main import MainController
from abstra_internals.usage import usage
from abstra_internals.utils import is_it_true


def get_editor_bp(controller: MainController):
    bp = flask.Blueprint("editor_hooks", __name__)

    @bp.route("/<path:id>", methods=["GET"])
    @usage
    def _get_hook(id: str):
        hook = controller.get_hook(id)
        if not hook:
            flask.abort(404)
        return hook.editor_dto

    @bp.route("/", methods=["GET"])
    @usage
    def _get_hooks():
        return [f.editor_dto for f in controller.get_hooks()]

    @bp.route("/", methods=["POST"])
    @usage
    def _create_hook():
        data = flask.request.json
        if not data:
            flask.abort(400)
        title = data.get("title")
        file = data.get("file")
        if not title or not file:
            flask.abort(400)
        workflow_position = data.get("position", (0, 0))
        hook = controller.create_hook(title, file, workflow_position)
        return hook.editor_dto

    @bp.route("/<path:id>", methods=["PUT"])
    @usage
    def _update_hook(id: str):
        changes = flask.request.json
        if not changes:
            flask.abort(400)

        hook = controller.update_stage(id, changes)
        return hook.editor_dto if hook else None

    @bp.route("/<path:id>", methods=["DELETE"])
    @usage
    def _delete_hook(id: str):
        remove_file = flask.request.args.get(
            "remove_file", default=False, type=is_it_true
        )
        controller.delete_hook(id, remove_file)
        return {"success": True}

    @bp.route("/<path:id>/run", methods=["POST", "GET", "PUT", "DELETE", "PATCH"])
    @usage
    def _run_hook(id: str):
        hook = controller.get_hook(id)
        if not hook:
            flask.abort(404)

        request_context = context_from_flask(flask.request)

        client = HookClient(request_context)

        ExecutionController(
            stage=hook,
            client=client,
            request=request_context,
            workflow_engine=controller.workflow_engine,
            stage_run_repository=controller.stage_run_repository,
            execution_repository=controller.execution_repository,
            target_stage_run_id=flask.request.args.get(STAGE_RUN_ID_PARAM_KEY),
        ).run().wait()

        return {
            "body": client.response["body"],
            "status": client.response["status"],
            "headers": client.response["headers"],
        }

    @bp.route("/<path:id>/test", methods=["POST", "GET", "PUT", "DELETE", "PATCH"])
    @usage
    def _test_hook(id: str):
        hook = controller.get_hook(id)
        if not hook:
            flask.abort(404)

        request_context = context_from_flask(flask.request)

        client = HookClient(request_context)

        DetachedExecutionController(
            stage=hook,
            client=client,
            request=request_context,
            target_stage_run_id=None,
            workflow_engine=controller.workflow_engine,
            stage_run_repository=controller.stage_run_repository,
            execution_repository=controller.execution_repository,
        ).run().wait()

        return {
            "body": client.response["body"],
            "status": client.response["status"],
            "headers": client.response["headers"],
        }

    return bp



import abstra.forms as af
import abstra.workflows as aw

# Stage Title: Employee Information Collection
# Collect employee information
name = af.read("👋 Welcome! What is your full name?")
email = af.read_email("📧 Please enter your company email:")
department = af.read_dropdown(
    "🏢 Select your department:",
    options=["Product", "Engineering", "Marketing", "Data", "Operations"],
)

# Store data in the current thread
aw.set_data("name", name)
aw.set_data("email", email)
aw.set_data("department", department)

# Display a welcome message
af.display(f"🎉 Welcome to the team, {name}!")



import flask

from abstra_internals.server.controller.main import MainController
from abstra_internals.settings import Settings
from abstra_internals.templates import (
    new_form_code,
    new_hook_code,
    new_job_code,
    new_script_code,
)
from abstra_internals.usage import usage


def get_editor_bp(controller: MainController):
    bp = flask.Blueprint("editor_workspace", __name__)

    @bp.get("/")
    @usage
    def _get_workspace():
        return controller.get_workspace().as_dict

    @bp.put("/")
    @usage
    def _update_workspace():
        if not flask.request.json:
            flask.abort(400)
        controller.update_workspace(flask.request.json)
        return controller.get_workspace().as_dict

    @bp.get("/root")
    @usage
    def _get_workspace_root_path():
        return str(Settings.root_path.absolute())

    @bp.post("/open-file")
    @usage
    def _open_file():
        if not flask.request.json:
            flask.abort(400)
        file_path = flask.request.json["path"]
        mode = flask.request.json.get("mode", "file")
        controller.open_file(file_path, create_if_not_exists=True, mode=mode)
        return {"success": True}

    @bp.post("/init-file")
    @usage
    def _init_file():
        if not flask.request.json:
            flask.abort(400)
        file_path = flask.request.json["path"]
        type = flask.request.json["type"]
        if type == "scripts":
            controller.init_code_file(file_path, new_script_code)
        elif type == "forms":
            controller.init_code_file(file_path, new_form_code)
        elif type == "hooks":
            controller.init_code_file(file_path, new_hook_code)
        elif type == "jobs":
            controller.init_code_file(file_path, new_job_code)
        else:
            flask.abort(400)
        return {"success": True}

    # 1s pooling in this route
    @bp.get("/read-file")
    def _read_file():
        file = flask.request.args.get("file")
        if not file:
            flask.abort(400)

        content = controller.read_file(file)
        if content is None:
            flask.abort(404)
        return content

    # 1s pooling in this route
    @bp.get("/check-file")
    def _check_file():
        file_path = flask.request.args.get("path")
        if not file_path:
            flask.abort(400)

        return {"exists": controller.check_file(file_path)}

    @bp.get("/read-test-data")
    @usage
    def _read_test_data():
        return controller.read_test_data()

    @bp.post("/write-test-data")
    @usage
    def _write_test_data():
        if not flask.request.json:
            flask.abort(400)
        test_data = flask.request.json.get("test_data", "{}")
        controller.write_test_data(test_data)
        return {"success": True}

    @bp.get("/files")
    @usage
    def _list_files():
        path = flask.request.args.get("path", ".")
        mode = flask.request.args.get("mode", "file")
        return controller.list_files(path, mode)

    @bp.post("/deploy")
    @usage
    def _deploy():
        controller.deploy()
        return {"success": True}

    return bp



import abc

from abstra_internals.entities.execution import Execution


class IWorkflowEngine(abc.ABC):
    @abc.abstractmethod
    def handle_execution_end(self, execution: Execution):
        raise NotImplementedError()



from typing import List, Union

from abstra_internals.email_templates import message_template
from abstra_internals.repositories import email_repository


def send_email(to: Union[str, List[str]], message: str, title: str = ""):
    if isinstance(to, str):
        to = [to]

    email_repository.send(message_template.generate_email(to, message, title))


__all__ = [
    "send_email",
]



import typing
from dataclasses import dataclass

import jwt

from abstra_internals.environment import PROJECT_ID, PUBLIC_KEY
from abstra_internals.utils.email import is_valid_email

USER_AUTH_HEADER_KEY = "Authorization"
API_AUTH_HEADER_KEY = "Api-Authorization"
AUTHOR_AUTH_HEADER_KEY = "Author-Authorization"


def decode_jwt(jwt_str: str):
    try:
        if PUBLIC_KEY:
            return jwt.decode(
                jwt_str, key=PUBLIC_KEY, algorithms=["RS256"], audience=PROJECT_ID
            )
        return jwt.decode(
            jwt_str, options={"verify_signature": False}, audience=PROJECT_ID
        )

    except Exception as e:
        print("error decoding jwt", e)
        return None


@dataclass
class UserClaims:
    """The response from the authentication process

    Attributes:
      email (str): The email address of the user
      claims (dict): The claims from the JWT token
    """

    claims: typing.Dict[str, typing.Any]

    @property
    def email(self) -> str:
        return self.claims["email"]

    @classmethod
    def from_jwt(cls, jwt_str: str) -> typing.Optional["UserClaims"]:
        claims = decode_jwt(jwt_str)
        if claims is None:
            return None

        email = claims.get("email")
        if not is_valid_email(email):
            return None

        return cls(claims)

    def __getattr__(self, name):
        return self.claims.get(name)



from abstra_internals.repositories.project.json_migrations.base_migration import (
    Migration,
)


class Migration005(Migration):
    @staticmethod
    def target_version() -> str:
        return "5.0"

    @staticmethod
    def source_version() -> str:
        return "4.0"

    def _migrate(self) -> None:
        conditions = self.data.get("conditions", [])

        for condition in conditions:
            if "title" in condition:
                del condition["title"]

        iterators = self.data.get("iterators", [])
        for iterator in iterators:
            if "title" in iterator:
                del iterator["title"]



import json
from threading import Thread
from typing import Optional

from abstra_internals.controllers.execution_client import ExecutionClient
from abstra_internals.controllers.execution_target import ExecutionTarget
from abstra_internals.controllers.workflow_engine_detached import DetachedWorkflowEngine
from abstra_internals.controllers.workflow_interface import IWorkflowEngine
from abstra_internals.entities.execution import Execution, RequestContext
from abstra_internals.repositories.execution import ExecutionRepository
from abstra_internals.repositories.project.project import ActionStage, ProjectRepository
from abstra_internals.repositories.stage_run import StageRunRepository
from abstra_internals.settings import Settings
from abstra_internals.utils.dot_abstra import TEST_DATA_FILE


class LockFailedException(Exception):
    pass


class UnsetThreadException(Exception):
    pass


class NotStartedException(Exception):
    pass


class ExecutionController:
    pthread: Optional[Thread] = None

    def __init__(
        self,
        *,
        stage: ActionStage,
        client: ExecutionClient,
        workflow_engine: IWorkflowEngine,
        request: Optional[RequestContext],
        target_stage_run_id: Optional[str],
        stage_run_repository: StageRunRepository,
        execution_repository: ExecutionRepository,
    ) -> None:
        self.stage = stage
        self.client = client
        self.request = request
        self.workflow_engine = workflow_engine
        self.target_stage_run_id = target_stage_run_id
        self.stage_run_repository = stage_run_repository
        self.execution_repository = execution_repository

    def run(self):
        if not self.target_stage_run_id:
            if not ProjectRepository.load().is_initial(self.stage):
                self.client.handle_unset_thread()
                raise UnsetThreadException()

            self.target_stage_run_id = self.stage_run_repository.create_initial(
                self.stage.id
            ).id

        self.target_stage_run_id = self.stage_run_repository.ensure_not_abandoned(
            self.target_stage_run_id
        )

        execution = Execution.create(
            stage_run_id=self.target_stage_run_id,
            request_context=self.request,
            stage_id=self.stage.id,
        )

        if not self.stage_run_repository.acquire_lock(
            stage_run_id=self.target_stage_run_id, execution_id=execution.id
        ):
            self.client.handle_lock_failed(self.target_stage_run_id)
            raise LockFailedException()

        self.pthread = Thread(
            target=ExecutionTarget,
            kwargs=dict(
                stage=self.stage,
                client=self.client,
                execution=execution,
                workflow_engine=self.workflow_engine,
                stage_run_repository=self.stage_run_repository,
                execution_repository=self.execution_repository,
            ),
            name=f"{self.stage.title} - {execution.short_id}",
        )

        self.pthread.start()
        return self

    def wait(self):
        if not self.pthread:
            raise NotStartedException()

        self.pthread.join()


class DetachedExecutionController(ExecutionController):
    @staticmethod
    def read_test_data() -> dict:
        test_file = Settings.root_path / TEST_DATA_FILE
        if not test_file.is_file():
            return {}

        return json.loads(test_file.read_text(encoding="utf-8"))

    def __init__(
        self,
        *,
        stage: ActionStage,
        client: ExecutionClient,
        request: Optional[RequestContext],
        stage_run_repository: StageRunRepository,
        execution_repository: ExecutionRepository,
        workflow_engine: IWorkflowEngine,
        target_stage_run_id: Optional[str],
    ) -> None:
        detached_stage_run = stage_run_repository.create_detached(
            thread_data=self.read_test_data(),
            stage_id=stage.id,
        )

        super().__init__(
            stage=stage,
            client=client,
            request=request,
            target_stage_run_id=detached_stage_run.id,
            stage_run_repository=stage_run_repository,
            execution_repository=execution_repository,
            workflow_engine=DetachedWorkflowEngine(stage_run_repository),
        )



import os
from typing import List

from dotenv import load_dotenv

from abstra_internals.linter.linter import LinterFix, LinterIssue, LinterRule
from abstra_internals.repositories.env_vars import EnvVarsRepository
from abstra_internals.settings import Settings


class LoadEnvVars(LinterFix):
    label = "Load environment variables"

    def fix(self):
        load_dotenv(Settings.root_path / ".env", override=True)


class EnvFileChangeFound(LinterIssue):
    def __init__(self) -> None:
        self.label = "Your .env file was updated. Please load the new environment variables to ensure consistency."
        self.fixes = [LoadEnvVars()]


class EnvFileChanged(LinterRule):
    label = "Changes to .env file detected"
    type = "bug"
    initial_file_exists = None

    def find_issues(self) -> List[LinterIssue]:
        env_file = Settings.root_path / ".env"
        current_file_exists = env_file.exists()

        if EnvFileChanged.initial_file_exists is None:
            EnvFileChanged.initial_file_exists = current_file_exists
            if current_file_exists:
                return []

        if EnvFileChanged.initial_file_exists != current_file_exists:
            return [EnvFileChangeFound()]

        env_vars = EnvVarsRepository.list()
        for env in env_vars:
            if env.value != os.getenv(env.name):
                return [EnvFileChangeFound()]

        return []
